@@ -1,3 +1,27 @@
+/*
+ * Copyright (C) 2006 Michael Brown <mbrown@fensystems.co.uk>.
+ * Copyright (C) 2006 Nikhil Chandru Rao
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+ * 02110-1301, USA.
+ *
+ * You can also choose to distribute this program under the terms of
+ * the Unmodified Binary Distribution Licence (as given in the file
+ * COPYING.UBDL), provided that you have satisfied its requirements.
+ */
+
 #include <string.h>
 #include <stdint.h>
 #include <stdlib.h>
@@ -14,7 +38,9 @@
 #include <ipxe/tcpip.h>
 #include <ipxe/dhcp.h>
 #include <ipxe/settings.h>
-#include <ipxe/timer.h>
+#include <ipxe/fragment.h>
+#include <ipxe/ipstat.h>
+#include <ipxe/profile.h>
 
 /** @file
  *
@@ -22,7 +48,7 @@
  *
  */
 
-FILE_LICENCE ( GPL2_OR_LATER );
+FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );
 
 /* Unique IP datagram identification number (high byte) */
 static uint8_t next_ident_high = 0;
@@ -30,11 +56,21 @@ static uint8_t next_ident_high = 0;
 /** List of IPv4 miniroutes */
 struct list_head ipv4_miniroutes = LIST_HEAD_INIT ( ipv4_miniroutes );
 
-/** List of fragment reassembly buffers */
-static LIST_HEAD ( ipv4_fragments );
+/** IPv4 statistics */
+static struct ip_statistics ipv4_stats;
 
-/** Fragment reassembly timeout */
-#define IP_FRAG_TIMEOUT ( TICKS_PER_SEC / 2 )
+/** IPv4 statistics family */
+struct ip_statistics_family
+ipv4_stats_family __ip_statistics_family ( IP_STATISTICS_IPV4 ) = {
+	.version = 4,
+	.stats = &ipv4_stats,
+};
+
+/** Transmit profiler */
+static struct profiler ipv4_tx_profiler __profiler = { .name = "ipv4.tx" };
+
+/** Receive profiler */
+static struct profiler ipv4_rx_profiler __profiler = { .name = "ipv4.rx" };
 
 /**
  * Add IPv4 minirouting table entry
@@ -103,6 +139,7 @@ static void del_ipv4_miniroute ( struct ipv4_miniroute *miniroute ) {
 /**
  * Perform IPv4 routing
  *
+ * @v scope_id		Destination address scope ID
  * @v dest		Final destination address
  * @ret dest		Next hop destination address
  * @ret miniroute	Routing table entry to use, or NULL if no route
@@ -110,22 +147,42 @@ static void del_ipv4_miniroute ( struct ipv4_miniroute *miniroute ) {
  * If the route requires use of a gateway, the next hop destination
  * address will be overwritten with the gateway address.
  */
-static struct ipv4_miniroute * ipv4_route ( struct in_addr *dest ) {
+static struct ipv4_miniroute * ipv4_route ( unsigned int scope_id,
+					    struct in_addr *dest ) {
 	struct ipv4_miniroute *miniroute;
-	int local;
-	int has_gw;
 
 	/* Find first usable route in routing table */
 	list_for_each_entry ( miniroute, &ipv4_miniroutes, list ) {
+
+		/* Skip closed network devices */
 		if ( ! netdev_is_open ( miniroute->netdev ) )
 			continue;
-		local = ( ( ( dest->s_addr ^ miniroute->address.s_addr )
-			    & miniroute->netmask.s_addr ) == 0 );
-		has_gw = ( miniroute->gateway.s_addr );
-		if ( local || has_gw ) {
-			if ( ! local )
+
+		if ( IN_IS_MULTICAST ( dest->s_addr ) ) {
+
+			/* If destination is non-global, and the scope ID
+			 * matches this network device, then use this route.
+			 */
+			if ( miniroute->netdev->index == scope_id )
+				return miniroute;
+
+		} else {
+
+			/* If destination is an on-link global
+			 * address, then use this route.
+			 */
+			if ( ( ( dest->s_addr ^ miniroute->address.s_addr )
+			       & miniroute->netmask.s_addr ) == 0 )
+				return miniroute;
+
+			/* If destination is an off-link global
+			 * address, and we have a default gateway,
+			 * then use this route.
+			 */
+			if ( miniroute->gateway.s_addr ) {
 				*dest = miniroute->gateway;
-			return miniroute;
+				return miniroute;
+			}
 		}
 	}
 
@@ -133,131 +190,79 @@ static struct ipv4_miniroute * ipv4_route ( struct in_addr *dest ) {
 }
 
 /**
- * Expire fragment reassembly buffer
+ * Determine transmitting network device
  *
- * @v timer		Retry timer
- * @v fail		Failure indicator
+ * @v st_dest		Destination network-layer address
+ * @ret netdev		Transmitting network device, or NULL
  */
-static void ipv4_fragment_expired ( struct retry_timer *timer,
-				    int fail __unused ) {
-	struct ipv4_fragment *frag =
-		container_of ( timer, struct ipv4_fragment, timer );
-	struct iphdr *iphdr = frag->iobuf->data;
+static struct net_device * ipv4_netdev ( struct sockaddr_tcpip *st_dest ) {
+	struct sockaddr_in *sin_dest = ( ( struct sockaddr_in * ) st_dest );
+	struct in_addr dest = sin_dest->sin_addr;
+	struct ipv4_miniroute *miniroute;
 
-	DBGC ( iphdr->src, "IPv4 fragment %04x expired\n",
-	       ntohs ( iphdr->ident ) );
-	free_iob ( frag->iobuf );
-	list_del ( &frag->list );
-	free ( frag );
+	/* Find routing table entry */
+	miniroute = ipv4_route ( sin_dest->sin_scope_id, &dest );
+	if ( ! miniroute )
+		return NULL;
+
+	return miniroute->netdev;
 }
 
 /**
- * Find matching fragment reassembly buffer
+ * Check if IPv4 fragment matches fragment reassembly buffer
  *
- * @v iphdr		IPv4 header
- * @ret frag		Fragment reassembly buffer, or NULL
+ * @v fragment		Fragment reassembly buffer
+ * @v iobuf		I/O buffer
+ * @v hdrlen		Length of non-fragmentable potion of I/O buffer
+ * @ret is_fragment	Fragment matches this reassembly buffer
  */
-static struct ipv4_fragment * ipv4_fragment ( struct iphdr *iphdr ) {
-	struct ipv4_fragment *frag;
-	struct iphdr *frag_iphdr;
+static int ipv4_is_fragment ( struct fragment *fragment,
+			      struct io_buffer *iobuf,
+			      size_t hdrlen __unused ) {
+	struct iphdr *frag_iphdr = fragment->iobuf->data;
+	struct iphdr *iphdr = iobuf->data;
 
-	list_for_each_entry ( frag, &ipv4_fragments, list ) {
-		frag_iphdr = frag->iobuf->data;
-
-		if ( ( iphdr->src.s_addr == frag_iphdr->src.s_addr ) &&
-		     ( iphdr->ident == frag_iphdr->ident ) ) {
-			return frag;
-		}
-	}
-
-	return NULL;
+	return ( ( iphdr->src.s_addr == frag_iphdr->src.s_addr ) &&
+		 ( iphdr->ident == frag_iphdr->ident ) );
 }
 
 /**
- * Fragment reassembler
+ * Get IPv4 fragment offset
  *
  * @v iobuf		I/O buffer
- * @ret iobuf		Reassembled packet, or NULL
+ * @v hdrlen		Length of non-fragmentable potion of I/O buffer
+ * @ret offset		Offset
  */
-static struct io_buffer * ipv4_reassemble ( struct io_buffer *iobuf ) {
+static size_t ipv4_fragment_offset ( struct io_buffer *iobuf,
+				     size_t hdrlen __unused ) {
 	struct iphdr *iphdr = iobuf->data;
-	size_t offset = ( ( ntohs ( iphdr->frags ) & IP_MASK_OFFSET ) << 3 );
-	unsigned int more_frags = ( iphdr->frags & htons ( IP_MASK_MOREFRAGS ));
-	size_t hdrlen = ( ( iphdr->verhdrlen & IP_MASK_HLEN ) * 4 );
-	struct ipv4_fragment *frag;
-	size_t expected_offset;
-	struct io_buffer *new_iobuf;
 
-	/* Find matching fragment reassembly buffer, if any */
-	frag = ipv4_fragment ( iphdr );
-
-	/* Drop out-of-order fragments */
-	expected_offset = ( frag ? frag->offset : 0 );
-	if ( offset != expected_offset ) {
-		DBGC ( iphdr->src, "IPv4 dropping out-of-sequence fragment "
-		       "%04x (%zd+%zd, expected %zd)\n",
-		       ntohs ( iphdr->ident ), offset,
-		      ( iob_len ( iobuf ) - hdrlen ), expected_offset );
-		goto drop;
-	}
-
-	/* Create or extend fragment reassembly buffer as applicable */
-	if ( frag == NULL ) {
-
-		/* Create new fragment reassembly buffer */
-		frag = zalloc ( sizeof ( *frag ) );
-		if ( ! frag )
-			goto drop;
-		list_add ( &frag->list, &ipv4_fragments );
-		frag->iobuf = iobuf;
-		frag->offset = ( iob_len ( iobuf ) - hdrlen );
-		timer_init ( &frag->timer, ipv4_fragment_expired, NULL );
-
-	} else {
-
-		/* Extend reassembly buffer */
-		iob_pull ( iobuf, hdrlen );
-		new_iobuf = alloc_iob ( iob_len ( frag->iobuf ) +
-					iob_len ( iobuf ) );
-		if ( ! new_iobuf ) {
-			DBGC ( iphdr->src, "IPv4 could not extend reassembly "
-			       "buffer to %zd bytes\n",
-			       iob_len ( frag->iobuf ) + iob_len ( iobuf ) );
-			goto drop;
-		}
-		memcpy ( iob_put ( new_iobuf, iob_len ( frag->iobuf ) ),
-			 frag->iobuf->data, iob_len ( frag->iobuf ) );
-		memcpy ( iob_put ( new_iobuf, iob_len ( iobuf ) ),
-			 iobuf->data, iob_len ( iobuf ) );
-		free_iob ( frag->iobuf );
-		frag->iobuf = new_iobuf;
-		frag->offset += iob_len ( iobuf );
-		free_iob ( iobuf );
-		iphdr = frag->iobuf->data;
-		iphdr->len = ntohs ( iob_len ( frag->iobuf ) );
-
-		/* Stop fragment reassembly timer */
-		stop_timer ( &frag->timer );
-
-		/* If this is the final fragment, return it */
-		if ( ! more_frags ) {
-			iobuf = frag->iobuf;
-			list_del ( &frag->list );
-			free ( frag );
-			return iobuf;
-		}
-	}
-
-	/* (Re)start fragment reassembly timer */
-	start_timer_fixed ( &frag->timer, IP_FRAG_TIMEOUT );
-
-	return NULL;
-
- drop:
-	free_iob ( iobuf );
-	return NULL;
+	return ( ( ntohs ( iphdr->frags ) & IP_MASK_OFFSET ) << 3 );
 }
 
+/**
+ * Check if more fragments exist
+ *
+ * @v iobuf		I/O buffer
+ * @v hdrlen		Length of non-fragmentable potion of I/O buffer
+ * @ret more_frags	More fragments exist
+ */
+static int ipv4_more_fragments ( struct io_buffer *iobuf,
+				 size_t hdrlen __unused ) {
+	struct iphdr *iphdr = iobuf->data;
+
+	return ( iphdr->frags & htons ( IP_MASK_MOREFRAGS ) );
+}
+
+/** IPv4 fragment reassembler */
+static struct fragment_reassembler ipv4_reassembler = {
+	.list = LIST_HEAD_INIT ( ipv4_reassembler.list ),
+	.is_fragment = ipv4_is_fragment,
+	.fragment_offset = ipv4_fragment_offset,
+	.more_fragments = ipv4_more_fragments,
+	.stats = &ipv4_stats,
+};
+
 /**
  * Add IPv4 pseudo-header checksum to existing checksum
  *
@@ -310,6 +315,12 @@ static int ipv4_tx ( struct io_buffer *iobuf,
 	const void *ll_dest;
 	int rc;
 
+	/* Start profiling */
+	profile_start ( &ipv4_tx_profiler );
+
+	/* Update statistics */
+	ipv4_stats.out_requests++;
+
 	/* Fill up the IP header, except source address */
 	memset ( iphdr, 0, sizeof ( *iphdr ) );
 	iphdr->verhdrlen = ( IP_VER | ( sizeof ( *iphdr ) / 4 ) );
@@ -324,8 +335,8 @@ static int ipv4_tx ( struct io_buffer *iobuf,
 	if ( sin_src )
 		iphdr->src = sin_src->sin_addr;
 	if ( ( next_hop.s_addr != INADDR_BROADCAST ) &&
-	     ( ! IN_MULTICAST ( ntohl ( next_hop.s_addr ) ) ) &&
-	     ( ( miniroute = ipv4_route ( &next_hop ) ) != NULL ) ) {
+	     ( ( miniroute = ipv4_route ( sin_dest->sin_scope_id,
+					  &next_hop ) ) != NULL ) ) {
 		iphdr->src = miniroute->address;
 		netmask = miniroute->netmask;
 		netdev = miniroute->netdev;
@@ -333,6 +344,7 @@ static int ipv4_tx ( struct io_buffer *iobuf,
 	if ( ! netdev ) {
 		DBGC ( sin_dest->sin_addr, "IPv4 has no route to %s\n",
 		       inet_ntoa ( iphdr->dest ) );
+		ipv4_stats.out_no_routes++;
 		rc = -ENETUNREACH;
 		goto err;
 	}
@@ -360,15 +372,17 @@ static int ipv4_tx ( struct io_buffer *iobuf,
 	/* Calculate link-layer destination address, if possible */
 	if ( ( ( next_hop.s_addr ^ INADDR_BROADCAST ) & ~netmask.s_addr ) == 0){
 		/* Broadcast address */
+		ipv4_stats.out_bcast_pkts++;
 		ll_dest = netdev->ll_broadcast;
-	} else if ( IN_MULTICAST ( ntohl ( next_hop.s_addr ) ) ) {
+	} else if ( IN_IS_MULTICAST ( next_hop.s_addr ) ) {
 		/* Multicast address */
+		ipv4_stats.out_mcast_pkts++;
 		if ( ( rc = netdev->ll_protocol->mc_hash ( AF_INET, &next_hop,
 							   ll_dest_buf ) ) !=0){
 			DBGC ( sin_dest->sin_addr, "IPv4 could not hash "
 			       "multicast %s: %s\n",
 			       inet_ntoa ( next_hop ), strerror ( rc ) );
-			return rc;
+			goto err;
 		}
 		ll_dest = ll_dest_buf;
 	} else {
@@ -376,6 +390,10 @@ static int ipv4_tx ( struct io_buffer *iobuf,
 		ll_dest = NULL;
 	}
 
+	/* Update statistics */
+	ipv4_stats.out_transmits++;
+	ipv4_stats.out_octets += iob_len ( iobuf );
+
 	/* Hand off to link layer (via ARP if applicable) */
 	if ( ll_dest ) {
 		if ( ( rc = net_tx ( iobuf, netdev, &ipv4_protocol, ll_dest,
@@ -395,6 +413,7 @@ static int ipv4_tx ( struct io_buffer *iobuf,
 		}
 	}
 
+	profile_stop ( &ipv4_tx_profiler );
 	return 0;
 
  err:
@@ -467,43 +486,56 @@ static int ipv4_rx ( struct io_buffer *iobuf,
 	uint16_t pshdr_csum;
 	int rc;
 
+	/* Start profiling */
+	profile_start ( &ipv4_rx_profiler );
+
+	/* Update statistics */
+	ipv4_stats.in_receives++;
+	ipv4_stats.in_octets += iob_len ( iobuf );
+	if ( flags & LL_BROADCAST ) {
+		ipv4_stats.in_bcast_pkts++;
+	} else if ( flags & LL_MULTICAST ) {
+		ipv4_stats.in_mcast_pkts++;
+	}
+
 	/* Sanity check the IPv4 header */
 	if ( iob_len ( iobuf ) < sizeof ( *iphdr ) ) {
 		DBGC ( iphdr->src, "IPv4 packet too short at %zd bytes (min "
 		       "%zd bytes)\n", iob_len ( iobuf ), sizeof ( *iphdr ) );
-		goto err;
+		goto err_header;
 	}
 	if ( ( iphdr->verhdrlen & IP_MASK_VER ) != IP_VER ) {
 		DBGC ( iphdr->src, "IPv4 version %#02x not supported\n",
 		       iphdr->verhdrlen );
-		goto err;
+		goto err_header;
 	}
 	hdrlen = ( ( iphdr->verhdrlen & IP_MASK_HLEN ) * 4 );
 	if ( hdrlen < sizeof ( *iphdr ) ) {
 		DBGC ( iphdr->src, "IPv4 header too short at %zd bytes (min "
 		       "%zd bytes)\n", hdrlen, sizeof ( *iphdr ) );
-		goto err;
+		goto err_header;
 	}
 	if ( hdrlen > iob_len ( iobuf ) ) {
 		DBGC ( iphdr->src, "IPv4 header too long at %zd bytes "
 		       "(packet is %zd bytes)\n", hdrlen, iob_len ( iobuf ) );
-		goto err;
+		goto err_header;
 	}
 	if ( ( csum = tcpip_chksum ( iphdr, hdrlen ) ) != 0 ) {
 		DBGC ( iphdr->src, "IPv4 checksum incorrect (is %04x "
 		       "including checksum field, should be 0000)\n", csum );
-		goto err;
+		goto err_header;
 	}
 	len = ntohs ( iphdr->len );
 	if ( len < hdrlen ) {
 		DBGC ( iphdr->src, "IPv4 length too short at %zd bytes "
 		       "(header is %zd bytes)\n", len, hdrlen );
-		goto err;
+		goto err_header;
 	}
 	if ( len > iob_len ( iobuf ) ) {
 		DBGC ( iphdr->src, "IPv4 length too long at %zd bytes "
 		       "(packet is %zd bytes)\n", len, iob_len ( iobuf ) );
-		goto err;
+		ipv4_stats.in_truncated_pkts++;
+		goto err_other;
 	}
 
 	/* Truncate packet to correct length */
@@ -521,19 +553,20 @@ static int ipv4_rx ( struct io_buffer *iobuf,
 	     ( ! ipv4_has_addr ( netdev, iphdr->dest ) ) ) {
 		DBGC ( iphdr->src, "IPv4 discarding non-local unicast packet "
 		       "for %s\n", inet_ntoa ( iphdr->dest ) );
-		goto err;
+		ipv4_stats.in_addr_errors++;
+		goto err_other;
 	}
 
 	/* Perform fragment reassembly if applicable */
 	if ( iphdr->frags & htons ( IP_MASK_OFFSET | IP_MASK_MOREFRAGS ) ) {
-		/* Pass the fragment to ipv4_reassemble() which returns
+		/* Pass the fragment to fragment_reassemble() which returns
 		 * either a fully reassembled I/O buffer or NULL.
 		 */
-		iobuf = ipv4_reassemble ( iobuf );
+		iobuf = fragment_reassemble ( &ipv4_reassembler, iobuf,
+					      &hdrlen );
 		if ( ! iobuf )
 			return 0;
 		iphdr = iobuf->data;
-		hdrlen = ( ( iphdr->verhdrlen & IP_MASK_HLEN ) * 4 );
 	}
 
 	/* Construct socket addresses, calculate pseudo-header
@@ -547,16 +580,19 @@ static int ipv4_rx ( struct io_buffer *iobuf,
 	dest.sin.sin_addr = iphdr->dest;
 	pshdr_csum = ipv4_pshdr_chksum ( iobuf, TCPIP_EMPTY_CSUM );
 	iob_pull ( iobuf, hdrlen );
-	if ( ( rc = tcpip_rx ( iobuf, iphdr->protocol, &src.st,
-			       &dest.st, pshdr_csum ) ) != 0 ) {
+	if ( ( rc = tcpip_rx ( iobuf, netdev, iphdr->protocol, &src.st,
+			       &dest.st, pshdr_csum, &ipv4_stats ) ) != 0 ) {
 		DBGC ( src.sin.sin_addr, "IPv4 received packet rejected by "
 		       "stack: %s\n", strerror ( rc ) );
 		return rc;
 	}
 
+	profile_stop ( &ipv4_rx_profiler );
 	return 0;
 
- err:
+ err_header:
+	ipv4_stats.in_hdr_errors++;
+ err_other:
 	free_iob ( iobuf );
 	return -EINVAL;
 }
@@ -577,11 +613,43 @@ static int ipv4_arp_check ( struct net_device *netdev, const void *net_addr ) {
 	return -ENOENT;
 }
 
+/**
+ * Parse IPv4 address
+ *
+ * @v string		IPv4 address string
+ * @ret in		IPv4 address to fill in
+ * @ret ok		IPv4 address is valid
+ *
+ * Note that this function returns nonzero iff the address is valid,
+ * to match the standard BSD API function of the same name.  Unlike
+ * most other iPXE functions, a zero therefore indicates failure.
+ */
+int inet_aton ( const char *string, struct in_addr *in ) {
+	const char *separator = "...";
+	uint8_t *byte = ( ( uint8_t * ) in );
+	char *endp;
+	unsigned long value;
+
+	while ( 1 ) {
+		value = strtoul ( string, &endp, 0 );
+		if ( string == endp )
+			return 0;
+		if ( value > 0xff )
+			return 0;
+		*(byte++) = value;
+		if ( *endp != *separator )
+			return 0;
+		if ( ! *(separator++) )
+			return 1;
+		string = ( endp + 1 );
+	}
+}
+
 /**
  * Convert IPv4 address to dotted-quad notation
  *
- * @v in	IP address
- * @ret string	IP address in dotted-quad notation
+ * @v in		IPv4 address
+ * @ret string		IPv4 address in dotted-quad notation
  */
 char * inet_ntoa ( struct in_addr in ) {
 	static char buf[16]; /* "xxx.xxx.xxx.xxx" */
@@ -592,16 +660,46 @@ char * inet_ntoa ( struct in_addr in ) {
 }
 
 /**
- * Transcribe IP address
+ * Transcribe IPv4 address
  *
- * @v net_addr	IP address
- * @ret string	IP address in dotted-quad notation
+ * @v net_addr		IPv4 address
+ * @ret string		IPv4 address in dotted-quad notation
  *
  */
 static const char * ipv4_ntoa ( const void *net_addr ) {
 	return inet_ntoa ( * ( ( struct in_addr * ) net_addr ) );
 }
 
+/**
+ * Transcribe IPv4 socket address
+ *
+ * @v sa		Socket address
+ * @ret string		Socket address in standard notation
+ */
+static const char * ipv4_sock_ntoa ( struct sockaddr *sa ) {
+	struct sockaddr_in *sin = ( ( struct sockaddr_in * ) sa );
+
+	return inet_ntoa ( sin->sin_addr );
+}
+
+/**
+ * Parse IPv4 socket address
+ *
+ * @v string		Socket address string
+ * @v sa		Socket address to fill in
+ * @ret rc		Return status code
+ */
+static int ipv4_sock_aton ( const char *string, struct sockaddr *sa ) {
+	struct sockaddr_in *sin = ( ( struct sockaddr_in * ) sa );
+	struct in_addr in;
+
+	if ( inet_aton ( string, &in ) ) {
+		sin->sin_addr = in;
+		return 0;
+	}
+	return -EINVAL;
+}
+
 /** IPv4 protocol */
 struct net_protocol ipv4_protocol __net_protocol = {
 	.name = "IP",
@@ -615,7 +713,9 @@ struct net_protocol ipv4_protocol __net_protocol = {
 struct tcpip_net_protocol ipv4_tcpip_protocol __tcpip_net_protocol = {
 	.name = "IPv4",
 	.sa_family = AF_INET,
+	.header_len = sizeof ( struct iphdr ),
 	.tx = ipv4_tx,
+	.netdev = ipv4_netdev,
 };
 
 /** IPv4 ARP protocol */
@@ -624,6 +724,13 @@ struct arp_net_protocol ipv4_arp_protocol __arp_net_protocol = {
 	.check = ipv4_arp_check,
 };
 
+/** IPv4 socket address converter */
+struct sockaddr_converter ipv4_sockaddr_converter __sockaddr_converter = {
+	.family = AF_INET,
+	.ntoa = ipv4_sock_ntoa,
+	.aton = ipv4_sock_aton,
+};
+
 /******************************************************************************
  *
  * Settings
@@ -631,8 +738,53 @@ struct arp_net_protocol ipv4_arp_protocol __arp_net_protocol = {
  ******************************************************************************
  */
 
+/**
+ * Parse IPv4 address setting value
+ *
+ * @v type		Setting type
+ * @v value		Formatted setting value
+ * @v buf		Buffer to contain raw value
+ * @v len		Length of buffer
+ * @ret len		Length of raw value, or negative error
+ */
+int parse_ipv4_setting ( const struct setting_type *type __unused,
+			 const char *value, void *buf, size_t len ) {
+	struct in_addr ipv4;
+
+	/* Parse IPv4 address */
+	if ( inet_aton ( value, &ipv4 ) == 0 )
+		return -EINVAL;
+
+	/* Copy to buffer */
+	if ( len > sizeof ( ipv4 ) )
+		len = sizeof ( ipv4 );
+	memcpy ( buf, &ipv4, len );
+
+	return ( sizeof ( ipv4 ) );
+}
+
+/**
+ * Format IPv4 address setting value
+ *
+ * @v type		Setting type
+ * @v raw		Raw setting value
+ * @v raw_len		Length of raw setting value
+ * @v buf		Buffer to contain formatted value
+ * @v len		Length of buffer
+ * @ret len		Length of formatted value, or negative error
+ */
+int format_ipv4_setting ( const struct setting_type *type __unused,
+			  const void *raw, size_t raw_len, char *buf,
+			  size_t len ) {
+	const struct in_addr *ipv4 = raw;
+
+	if ( raw_len < sizeof ( *ipv4 ) )
+		return -EINVAL;
+	return snprintf ( buf, len, "%s", inet_ntoa ( *ipv4 ) );
+}
+
 /** IPv4 address setting */
-struct setting ip_setting __setting ( SETTING_IPv4 ) = {
+const struct setting ip_setting __setting ( SETTING_IP, ip ) = {
 	.name = "ip",
 	.description = "IP address",
 	.tag = DHCP_EB_YIADDR,
@@ -640,7 +792,7 @@ struct setting ip_setting __setting ( SETTING_IPv4 ) = {
 };
 
 /** IPv4 subnet mask setting */
-struct setting netmask_setting __setting ( SETTING_IPv4 ) = {
+const struct setting netmask_setting __setting ( SETTING_IP, netmask ) = {
 	.name = "netmask",
 	.description = "Subnet mask",
 	.tag = DHCP_SUBNET_MASK,
@@ -648,7 +800,7 @@ struct setting netmask_setting __setting ( SETTING_IPv4 ) = {
 };
 
 /** Default gateway setting */
-struct setting gateway_setting __setting ( SETTING_IPv4 ) = {
+const struct setting gateway_setting __setting ( SETTING_IP, gateway ) = {
 	.name = "gateway",
 	.description = "Default gateway",
 	.tag = DHCP_ROUTERS,
@@ -685,12 +837,12 @@ static int ipv4_create_routes ( void ) {
 		fetch_ipv4_setting ( settings, &netmask_setting, &netmask );
 		/* Calculate default netmask, if necessary */
 		if ( ! netmask.s_addr ) {
-			if ( IN_CLASSA ( ntohl ( address.s_addr ) ) ) {
-				netmask.s_addr = htonl ( IN_CLASSA_NET );
-			} else if ( IN_CLASSB ( ntohl ( address.s_addr ) ) ) {
-				netmask.s_addr = htonl ( IN_CLASSB_NET );
-			} else if ( IN_CLASSC ( ntohl ( address.s_addr ) ) ) {
-				netmask.s_addr = htonl ( IN_CLASSC_NET );
+			if ( IN_IS_CLASSA ( address.s_addr ) ) {
+				netmask.s_addr = INADDR_NET_CLASSA;
+			} else if ( IN_IS_CLASSB ( address.s_addr ) ) {
+				netmask.s_addr = INADDR_NET_CLASSB;
+			} else if ( IN_IS_CLASSC ( address.s_addr ) ) {
+				netmask.s_addr = INADDR_NET_CLASSC;
 			}
 		}
 		/* Get default gateway, if present */
@@ -710,5 +862,8 @@ struct settings_applicator ipv4_settings_applicator __settings_applicator = {
 	.apply = ipv4_create_routes,
 };
 
-/* Drag in ICMP */
-REQUIRE_OBJECT ( icmp );
+/* Drag in objects via ipv4_protocol */
+REQUIRING_SYMBOL ( ipv4_protocol );
+
+/* Drag in ICMPv4 */
+REQUIRE_OBJECT ( icmpv4 );
